@@ -460,12 +460,6 @@ _gnutls_io_read_buffered (gnutls_session_t session, size_t total,
       return GNUTLS_E_INVALID_REQUEST;
     }
 
-  if (ret < 0)
-    {
-      gnutls_assert ();
-      return ret;
-    }
-
   /* READ DATA
    */
   if (readsize > 0)
@@ -661,15 +655,14 @@ _gnutls_io_check_recv (gnutls_session_t session, unsigned int ms)
   reset_errno (session);
 
   ret = session->internals.pull_timeout_func(fd, ms);
+  if (ret == -1)
+    {
+      err = get_errno (session);
+      _gnutls_read_log ("READ_TIMEOUT: %d returned from %p, errno=%d (timeout: %u)\n",
+			(int) ret, fd, err, ms);
+      return errno_to_gerr(err);
+    }
 
-  err = get_errno (session);
-  if (ret == -1 && err == EINTR)
-    return GNUTLS_E_INTERRUPTED;
-  else if (ret == -1 && err == EAGAIN)
-    return GNUTLS_E_AGAIN;
-  else if (ret == -1)
-    return gnutls_assert_val(GNUTLS_E_PULL_ERROR);
-  
   if (ret > 0)
     return 0;
   else return GNUTLS_E_TIMEDOUT;
@@ -1177,7 +1170,6 @@ _gnutls_handshake_io_recv_int (gnutls_session_t session,
                                handshake_buffer_st * hsk, unsigned int optional)
 {
   int ret;
-  time_t tleft = 0;
 
   ret = get_last_packet(session, htype, hsk, optional);
   if (ret != GNUTLS_E_AGAIN && ret != GNUTLS_E_INTERRUPTED && ret != GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
