@@ -286,16 +286,26 @@ static const gnutls_error_entry error_algorithms[] = {
   ERROR_ENTRY (N_("Channel binding data not available"),
                GNUTLS_E_CHANNEL_BINDING_NOT_AVAILABLE, 1),
 
+  ERROR_ENTRY (N_("TPM error."),
+               GNUTLS_E_TPM_ERROR, 1),
+  ERROR_ENTRY (N_("TPM is not initialized."),
+               GNUTLS_E_TPM_UNINITIALIZED, 1),
+  ERROR_ENTRY (N_("TPM key was not found in persistent storage."),
+               GNUTLS_E_TPM_KEY_NOT_FOUND, 1),
+  ERROR_ENTRY (N_("Cannot initialize a session with the TPM."),
+               GNUTLS_E_TPM_SESSION_ERROR, 1),
   ERROR_ENTRY (N_("PKCS #11 error."),
                GNUTLS_E_PKCS11_ERROR, 1),
   ERROR_ENTRY (N_("PKCS #11 initialization error."),
                GNUTLS_E_PKCS11_LOAD_ERROR, 1),
   ERROR_ENTRY (N_("Error in parsing."),
                GNUTLS_E_PARSING_ERROR, 1),
-  ERROR_ENTRY (N_("PKCS #11 error in PIN."),
+  ERROR_ENTRY (N_("Error in provided PIN."),
                GNUTLS_E_PKCS11_PIN_ERROR, 1),
-  ERROR_ENTRY (N_("PKCS #11 PIN should be saved."),
-               GNUTLS_E_PKCS11_ERROR, 1),
+  ERROR_ENTRY (N_("Error in provided SRK password for TPM."),
+               GNUTLS_E_TPM_SRK_PASSWORD_ERROR, 1),
+  ERROR_ENTRY (N_("Error in provided password for key to be loaded in TPM."),
+               GNUTLS_E_TPM_KEY_PASSWORD_ERROR, 1),
   ERROR_ENTRY (N_("PKCS #11 error in slot"),
                GNUTLS_E_PKCS11_SLOT_ERROR, 1),
   ERROR_ENTRY (N_("Thread locking error"),
