@@ -78,8 +78,6 @@ void
 gnutls_record_disable_padding (gnutls_session_t session)
 {
   session->internals.priorities.no_padding = 1;
-  session->internals.priorities.allow_wrong_pms = 1;
-  session->internals.priorities.allow_large_records = 1;
 }
 
 /**
@@ -696,7 +694,7 @@ record_add_to_buffers (gnutls_session_t session,
                   ret = gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET);
                   goto unexpected_packet;
                 }
-
+                
               if (_dtls_is_async(session) && _dtls_async_timer_active(session))
                 {
                   if (session->security_parameters.entity == GNUTLS_SERVER &&
@@ -707,11 +705,9 @@ record_add_to_buffers (gnutls_session_t session,
                     }
                   else
                     {
-                      session->internals.recv_state = RECV_STATE_DTLS_RETRANSMIT;
                       ret = _dtls_retransmit(session);
                       if (ret == 0) 
                         {
-                          session->internals.recv_state = RECV_STATE_0;
                           ret = gnutls_assert_val(GNUTLS_E_AGAIN);
                           goto unexpected_packet;
                         }
@@ -865,7 +861,9 @@ record_read_headers (gnutls_session_t session,
 
 
 static int recv_headers( gnutls_session_t session, content_type_t type, 
-  gnutls_handshake_description_t htype, struct tls_record_st* record)
+                         gnutls_handshake_description_t htype, 
+                         struct tls_record_st* record,
+                         unsigned int ms)
 {
 int ret;
 gnutls_datum_t raw; /* raw headers */
@@ -874,7 +872,7 @@ gnutls_datum_t raw; /* raw headers */
   record->header_size = record->packet_size = RECORD_HEADER_SIZE(session);
 
   if ((ret =
-       _gnutls_io_read_buffered (session, record->header_size, -1)) != record->header_size)
+       _gnutls_io_read_buffered (session, record->header_size, -1, ms)) != record->header_size)
     {
       if (ret < 0 && gnutls_error_is_fatal (ret) == 0)
         return ret;
@@ -945,6 +943,9 @@ gnutls_datum_t raw; /* raw headers */
 
 /* This will receive record layer packets and add them to 
  * application_data_buffer and handshake_data_buffer.
+ *
+ * If the htype is not -1 then handshake timeouts
+ * will be enforced.
  */
 ssize_t
 _gnutls_recv_in_buffers (gnutls_session_t session, content_type_t type,
@@ -958,6 +959,7 @@ _gnutls_recv_in_buffers (gnutls_session_t session, content_type_t type,
   record_parameters_st *record_params;
   record_state_st *record_state;
   struct tls_record_st record;
+  time_t now, tleft = 0;
 
 begin:
 
@@ -990,8 +992,17 @@ begin:
 
   record_state = &record_params->read;
 
+  if (htype != (unsigned)-1 && session->internals.handshake_endtime > 0)
+    {
+      now = gnutls_time(0);
+      if (now < session->internals.handshake_endtime)
+        tleft = (session->internals.handshake_endtime - now) * 1000;
+      else
+        return gnutls_assert_val(GNUTLS_E_TIMEDOUT);
+    }
+
   /* receive headers */
-  ret = recv_headers(session, type, htype, &record);
+  ret = recv_headers(session, type, htype, &record, tleft);
   if (ret < 0)
     {
       ret = gnutls_assert_val_fatal(ret);
@@ -1003,11 +1014,20 @@ begin:
   else
     packet_sequence = &record_state->sequence_number;
 
+  if (htype != (unsigned)-1 && session->internals.handshake_endtime > 0)
+    {
+      now = gnutls_time(0);
+      if (now < session->internals.handshake_endtime)
+        tleft = (session->internals.handshake_endtime - now) * 1000;
+      else
+        return gnutls_assert_val(GNUTLS_E_TIMEDOUT);
+    }
+
   /* Read the packet data and insert it to record_recv_buffer.
    */
   ret =
        _gnutls_io_read_buffered (session, record.packet_size,
-                                 record.type);
+                                 record.type, tleft);
   if (ret != record.packet_size)
     {
       gnutls_assert();
@@ -1200,33 +1220,21 @@ _gnutls_recv_int (gnutls_session_t session, content_type_t type,
       gnutls_assert ();
       return GNUTLS_E_INVALID_SESSION;
     }
-    
-  switch(session->internals.recv_state)
-    {
-      case RECV_STATE_DTLS_RETRANSMIT:
-        ret = _dtls_retransmit(session);
-        if (ret < 0)
-          return gnutls_assert_val(ret);
-        
-        session->internals.recv_state = RECV_STATE_0;
-      case RECV_STATE_0:
 
-        _dtls_async_timer_check(session);
-        /* If we have enough data in the cache do not bother receiving
-         * a new packet. (in order to flush the cache)
-         */ 
-        ret = check_buffers (session, type, data, data_size, seq);
-        if (ret != 0)
-          return ret;
+  _dtls_async_timer_check(session);
+  
+  /* If we have enough data in the cache do not bother receiving
+   * a new packet. (in order to flush the cache)
+   */
+  ret = check_buffers (session, type, data, data_size, seq);
+  if (ret != 0)
+    return ret;
 
-        ret = _gnutls_recv_in_buffers(session, type, htype);
-        if (ret < 0 && ret != GNUTLS_E_SESSION_EOF)
-          return gnutls_assert_val(ret);
+  ret = _gnutls_recv_in_buffers(session, type, htype);
+  if (ret < 0 && ret != GNUTLS_E_SESSION_EOF)
+    return gnutls_assert_val(ret);
 
-        return check_buffers (session, type, data, data_size, seq);
-      default:
-        return gnutls_assert_val(GNUTLS_E_INTERNAL_ERROR);
-    }
+  return check_buffers (session, type, data, data_size, seq);
 }
 
 
