@@ -146,9 +146,9 @@ pkt_encode_len (cdk_stream_t out, size_t pktlen)
 {
   cdk_error_t rc;
 
-  if (!out)
-    return CDK_Inv_Value;
+  assert (out);
 
+  rc = 0;
   if (!pktlen)
     {
       /* Block mode, partial bodies, with 'DEF_BLOCKSIZE' from main.h */
@@ -179,8 +179,7 @@ write_head_new (cdk_stream_t out, size_t size, int type)
 {
   cdk_error_t rc;
 
-  if (!out)
-    return CDK_Inv_Value;
+  assert (out);
 
   if (type < 0 || type > 63)
     return CDK_Inv_Packet;
@@ -197,8 +196,7 @@ write_head_old (cdk_stream_t out, size_t size, int type)
   cdk_error_t rc;
   int ctb;
 
-  if (!out)
-    return CDK_Inv_Value;
+  assert (out);
 
   if (type < 0 || type > 16)
     return CDK_Inv_Packet;
@@ -259,8 +257,8 @@ write_pubkey_enc (cdk_stream_t out, cdk_pkt_pubkey_enc_t pke, int old_ctb)
   size_t size;
   int rc, nenc;
 
-  if (!out || !pke)
-    return CDK_Inv_Value;
+  assert (out);
+  assert (pke);
 
   if (pke->version < 2 || pke->version > 3)
     return CDK_Inv_Packet;
@@ -294,8 +292,8 @@ write_mdc (cdk_stream_t out, cdk_pkt_mdc_t mdc)
 {
   cdk_error_t rc;
 
-  if (!out || !mdc)
-    return CDK_Inv_Value;
+  assert (mdc);
+  assert (out);
 
   if (DEBUG_PKT)
     _gnutls_write_log ("write_mdc:\n");
@@ -365,85 +363,61 @@ write_signature (cdk_stream_t out, cdk_pkt_signature_t sig, int old_ctb)
   size_t nbytes, size, nsig;
   cdk_error_t rc;
 
-  if (!out || !sig)
-    return CDK_Inv_Value;
+  assert (out);
+  assert (sig);
 
   if (!KEY_CAN_SIGN (sig->pubkey_algo))
-    return gnutls_assert_val(CDK_Inv_Algo);
+    return CDK_Inv_Algo;
   if (sig->version < 2 || sig->version > 4)
-    return gnutls_assert_val(CDK_Inv_Packet);
+    return CDK_Inv_Packet;
 
   if (DEBUG_PKT)
     _gnutls_write_log ("write_signature:\n");
 
   nsig = cdk_pk_get_nsig (sig->pubkey_algo);
   if (!nsig)
-    return gnutls_assert_val(CDK_Inv_Algo);
+    return CDK_Inv_Algo;
   if (sig->version < 4)
     return write_v3_sig (out, sig, nsig);
 
   size = 10 + calc_subpktsize (sig->hashed)
     + calc_subpktsize (sig->unhashed) + calc_mpisize (sig->mpi, nsig);
-
   rc = pkt_write_head (out, 0, size, CDK_PKT_SIGNATURE);
-  if (rc)
-    return gnutls_assert_val(rc);
-
-  rc = stream_putc (out, 4);
-  if (rc)
-    return gnutls_assert_val(rc);
-
-  rc = stream_putc (out, sig->sig_class);
-  if (rc)
-    return gnutls_assert_val(rc);
-
-  rc = stream_putc (out, _cdk_pub_algo_to_pgp (sig->pubkey_algo));
-  if (rc)
-    return gnutls_assert_val(rc);
-
-  rc = stream_putc (out, _gnutls_hash_algo_to_pgp (sig->digest_algo));
-  if (rc)
-    return gnutls_assert_val(rc);
-
-  rc = write_16 (out, sig->hashed_size);
-  if (rc)
-    return gnutls_assert_val(rc);
-
-  buf = _cdk_subpkt_get_array (sig->hashed, 0, &nbytes);
-  if (!buf)
-    return gnutls_assert_val(CDK_Out_Of_Core);
-
-  rc = stream_write (out, buf, nbytes);
-  cdk_free (buf);
-  if (rc)
-    return gnutls_assert_val(rc);
-
-  rc = write_16 (out, sig->unhashed_size);
-  if (rc)
-    return gnutls_assert_val(rc);
-
-  buf = _cdk_subpkt_get_array (sig->unhashed, 0, &nbytes);
-  if (!buf)
-    return gnutls_assert_val(CDK_Out_Of_Core);
-
-  rc = stream_write (out, buf, nbytes);
-  cdk_free (buf);
-  if (rc)
-    return gnutls_assert_val(rc);
-
-  rc = stream_putc (out, sig->digest_start[0]);
-  if (rc)
-    return gnutls_assert_val(rc);
-
-  rc = stream_putc (out, sig->digest_start[1]);
-  if (rc)
-    return gnutls_assert_val(rc);
-
-  rc = write_mpibuf (out, sig->mpi, nsig);
-  if (rc)
-    return gnutls_assert_val(rc);
-
-  return 0;
+  if (!rc)
+    rc = stream_putc (out, 4);
+  if (!rc)
+    rc = stream_putc (out, sig->sig_class);
+  if (!rc)
+    rc = stream_putc (out, _cdk_pub_algo_to_pgp (sig->pubkey_algo));
+  if (!rc)
+    rc = stream_putc (out, _gnutls_hash_algo_to_pgp (sig->digest_algo));
+  if (!rc)
+    rc = write_16 (out, sig->hashed_size);
+  if (!rc)
+    {
+      buf = _cdk_subpkt_get_array (sig->hashed, 0, &nbytes);
+      if (!buf)
+        return CDK_Out_Of_Core;
+      rc = stream_write (out, buf, nbytes);
+      cdk_free (buf);
+    }
+  if (!rc)
+    rc = write_16 (out, sig->unhashed_size);
+  if (!rc)
+    {
+      buf = _cdk_subpkt_get_array (sig->unhashed, 0, &nbytes);
+      if (!buf)
+        return CDK_Out_Of_Core;
+      rc = stream_write (out, buf, nbytes);
+      cdk_free (buf);
+    }
+  if (!rc)
+    rc = stream_putc (out, sig->digest_start[0]);
+  if (!rc)
+    rc = stream_putc (out, sig->digest_start[1]);
+  if (!rc)
+    rc = write_mpibuf (out, sig->mpi, nsig);
+  return rc;
 }
 
 
@@ -455,8 +429,8 @@ write_public_key (cdk_stream_t out, cdk_pkt_pubkey_t pk,
   size_t npkey = 0, size = 6;
   cdk_error_t rc;
 
-  if (!out || !pk)
-    return CDK_Inv_Value;
+  assert (out);
+  assert (pk);
 
   if (pk->version < 2 || pk->version > 4)
     return CDK_Inv_Packet;
@@ -532,8 +506,8 @@ write_secret_key (cdk_stream_t out, cdk_pkt_seckey_t sk,
   int pkttype, s2k_mode;
   cdk_error_t rc;
 
-  if (!out || !sk)
-    return CDK_Inv_Value;
+  assert (out);
+  assert (sk);
 
   if (!sk->pk)
     return CDK_Inv_Value;
@@ -652,8 +626,8 @@ write_compressed (cdk_stream_t out, cdk_pkt_compressed_t cd)
 {
   cdk_error_t rc;
 
-  if (!out || !cd)
-    return CDK_Inv_Value;
+  assert (out);
+  assert (cd);
 
   if (DEBUG_PKT)
     _gnutls_write_log ("packet: write_compressed\n");
@@ -673,8 +647,8 @@ write_literal (cdk_stream_t out, cdk_pkt_literal_t pt, int old_ctb)
   size_t size;
   cdk_error_t rc;
 
-  if (!out || !pt)
-    return CDK_Inv_Value;
+  assert (out);
+  assert (pt);
 
   /* We consider a packet without a body as an invalid packet.
      At least one octet must be present. */
@@ -720,8 +694,8 @@ write_onepass_sig (cdk_stream_t out, cdk_pkt_onepass_sig_t sig)
 {
   cdk_error_t rc;
 
-  if (!out || !sig)
-    return CDK_Inv_Value;
+  assert (out);
+  assert (sig);
 
   if (sig->version != 3)
     return CDK_Inv_Packet;
@@ -800,9 +774,7 @@ cdk_pkt_write (cdk_stream_t out, cdk_packet_t pkt)
   if (!out || !pkt)
     return CDK_Inv_Value;
 
-  if (DEBUG_PKT)
-    _gnutls_write_log ("write packet pkttype=%d\n", pkt->pkttype);
-
+  _gnutls_write_log ("write packet pkttype=%d\n", pkt->pkttype);
   switch (pkt->pkttype)
     {
     case CDK_PKT_LITERAL:
