@@ -125,7 +125,7 @@ int
 _gnutls_openpgp_export (cdk_kbnode_t node,
                         gnutls_openpgp_crt_fmt_t format,
                         void *output_data,
-                        size_t * output_data_size, int priv)
+                        size_t * output_data_size, int private)
 {
   size_t input_data_size = *output_data_size;
   size_t calc_size;
@@ -141,7 +141,7 @@ _gnutls_openpgp_export (cdk_kbnode_t node,
 
   /* If the caller uses output_data == NULL then return what he expects.
    */
-  if (!output_data && format != GNUTLS_OPENPGP_FMT_BASE64)
+  if (!output_data)
     {
       gnutls_assert ();
       return GNUTLS_E_SHORT_MEMORY_BUFFER;
@@ -156,7 +156,7 @@ _gnutls_openpgp_export (cdk_kbnode_t node,
          buffer is large enough. */
       rc = cdk_armor_encode_buffer (in, *output_data_size,
                                     NULL, 0, &calc_size,
-                                    priv ? CDK_ARMOR_SECKEY :
+                                    private ? CDK_ARMOR_SECKEY :
                                     CDK_ARMOR_PUBKEY);
       if (rc || calc_size > input_data_size)
         {
@@ -168,7 +168,7 @@ _gnutls_openpgp_export (cdk_kbnode_t node,
 
       rc = cdk_armor_encode_buffer (in, *output_data_size,
                                     output_data, input_data_size, &calc_size,
-                                    priv ? CDK_ARMOR_SECKEY :
+                                    private ? CDK_ARMOR_SECKEY :
                                     CDK_ARMOR_PUBKEY);
       gnutls_free (in);
       *output_data_size = calc_size;
@@ -189,7 +189,7 @@ _gnutls_openpgp_export (cdk_kbnode_t node,
  * gnutls_openpgp_crt_export:
  * @key: Holds the key.
  * @format: One of gnutls_openpgp_crt_fmt_t elements.
- * @output_data: will contain the raw or base64 encoded key
+ * @output_data: will contain the key base64 encoded or raw
  * @output_data_size: holds the size of output_data (and will
  *   be replaced by the actual size of parameters)
  *
@@ -382,20 +382,18 @@ gnutls_openpgp_crt_get_pk_algorithm (gnutls_openpgp_crt_t key,
       int idx;
 
       idx = gnutls_openpgp_crt_get_subkey_idx (key, keyid);
-      if (idx != GNUTLS_OPENPGP_MASTER_KEYID_IDX)
-        {
-          algo =
-            gnutls_openpgp_crt_get_subkey_pk_algorithm (key, idx, bits);
-          return algo;
-        }
+      algo =
+        gnutls_openpgp_crt_get_subkey_pk_algorithm (key, idx, NULL);
     }
-
-  pkt = cdk_kbnode_find_packet (key->knode, CDK_PKT_PUBLIC_KEY);
-  if (pkt)
+  else
     {
-      if (bits)
-        *bits = cdk_pk_get_nbits (pkt->pkt.public_key);
-      algo = _gnutls_openpgp_get_algo (pkt->pkt.public_key->pubkey_algo);
+      pkt = cdk_kbnode_find_packet (key->knode, CDK_PKT_PUBLIC_KEY);
+      if (pkt)
+        {
+          if (bits)
+            *bits = cdk_pk_get_nbits (pkt->pkt.public_key);
+          algo = _gnutls_openpgp_get_algo (pkt->pkt.public_key->pubkey_algo);
+        }
     }
 
   return algo;
@@ -813,9 +811,6 @@ gnutls_openpgp_crt_get_subkey_revoked_status (gnutls_openpgp_crt_t key,
       return GNUTLS_E_INVALID_REQUEST;
     }
 
-  if (idx == GNUTLS_OPENPGP_MASTER_KEYID_IDX)
-    return gnutls_openpgp_crt_get_revoked_status(key);
-
   pkt = _get_public_subkey (key, idx);
   if (!pkt)
     return GNUTLS_E_OPENPGP_GETKEY_FAILED;
@@ -857,9 +852,6 @@ gnutls_openpgp_crt_get_subkey_pk_algorithm (gnutls_openpgp_crt_t key,
       return GNUTLS_PK_UNKNOWN;
     }
 
-  if (idx == GNUTLS_OPENPGP_MASTER_KEYID_IDX)
-    return gnutls_openpgp_crt_get_pk_algorithm(key, bits);
-
   pkt = _get_public_subkey (key, idx);
 
   algo = 0;
@@ -894,9 +886,6 @@ gnutls_openpgp_crt_get_subkey_creation_time (gnutls_openpgp_crt_t key,
   if (!key)
     return (time_t) - 1;
 
-  if (idx == GNUTLS_OPENPGP_MASTER_KEYID_IDX)
-    return gnutls_openpgp_crt_get_creation_time(key);
-
   pkt = _get_public_subkey (key, idx);
   if (pkt)
     timestamp = pkt->pkt.public_key->timestamp;
@@ -929,9 +918,6 @@ gnutls_openpgp_crt_get_subkey_expiration_time (gnutls_openpgp_crt_t key,
   if (!key)
     return (time_t) - 1;
 
-  if (idx == GNUTLS_OPENPGP_MASTER_KEYID_IDX)
-    return gnutls_openpgp_crt_get_expiration_time(key);
-
   pkt = _get_public_subkey (key, idx);
   if (pkt)
     expiredate = pkt->pkt.public_key->expiredate;
@@ -964,9 +950,6 @@ gnutls_openpgp_crt_get_subkey_id (gnutls_openpgp_crt_t key,
       gnutls_assert ();
       return GNUTLS_E_INVALID_REQUEST;
     }
-  
-  if (idx == GNUTLS_OPENPGP_MASTER_KEYID_IDX)
-    return gnutls_openpgp_crt_get_key_id(key, keyid);
 
   pkt = _get_public_subkey (key, idx);
   if (!pkt)
@@ -1006,9 +989,6 @@ gnutls_openpgp_crt_get_subkey_fingerprint (gnutls_openpgp_crt_t key,
       gnutls_assert ();
       return GNUTLS_E_INVALID_REQUEST;
     }
-  
-  if (idx == GNUTLS_OPENPGP_MASTER_KEYID_IDX)
-    return gnutls_openpgp_crt_get_fingerprint(key, fpr, fprlen);
 
   *fprlen = 0;
 
@@ -1044,7 +1024,6 @@ gnutls_openpgp_crt_get_subkey_idx (gnutls_openpgp_crt_t key,
 {
   int ret;
   uint32_t kid[2];
-  uint8_t master_id[GNUTLS_OPENPGP_KEYID_SIZE];
 
   if (!key)
     {
@@ -1052,12 +1031,6 @@ gnutls_openpgp_crt_get_subkey_idx (gnutls_openpgp_crt_t key,
       return GNUTLS_E_INVALID_REQUEST;
     }
 
-  ret = gnutls_openpgp_crt_get_key_id (key, master_id);
-  if (ret < 0)
-    return gnutls_assert_val(ret);  
-  if (memcmp(master_id, keyid, GNUTLS_OPENPGP_KEYID_SIZE)==0)
-    return GNUTLS_OPENPGP_MASTER_KEYID_IDX;
-
   KEYID_IMPORT (kid, keyid);
   ret = _gnutls_openpgp_find_subkey_idx (key->knode, kid, 0);
 
@@ -1098,9 +1071,6 @@ gnutls_openpgp_crt_get_subkey_usage (gnutls_openpgp_crt_t key,
       return GNUTLS_E_INVALID_REQUEST;
     }
 
-  if (idx == GNUTLS_OPENPGP_MASTER_KEYID_IDX)
-    return gnutls_openpgp_crt_get_key_usage(key, key_usage);
-
   pkt = _get_public_subkey (key, idx);
   if (!pkt)
     return GNUTLS_E_OPENPGP_SUBKEY_ERROR;
@@ -1498,9 +1468,6 @@ gnutls_openpgp_crt_get_subkey_pk_rsa_raw (gnutls_openpgp_crt_t crt,
   uint8_t keyid[GNUTLS_OPENPGP_KEYID_SIZE];
   int ret;
 
-  if (idx == GNUTLS_OPENPGP_MASTER_KEYID_IDX)
-    return gnutls_openpgp_crt_get_pk_rsa_raw(crt, m, e);
-
   ret = gnutls_openpgp_crt_get_subkey_id (crt, idx, keyid);
   if (ret < 0)
     {
@@ -1539,9 +1506,6 @@ gnutls_openpgp_crt_get_subkey_pk_dsa_raw (gnutls_openpgp_crt_t crt,
   uint8_t keyid[GNUTLS_OPENPGP_KEYID_SIZE];
   int ret;
 
-  if (idx == GNUTLS_OPENPGP_MASTER_KEYID_IDX)
-    return gnutls_openpgp_crt_get_pk_dsa_raw(crt, p,q, g, y);
-
   ret = gnutls_openpgp_crt_get_subkey_id (crt, idx, keyid);
   if (ret < 0)
     {
@@ -1588,9 +1552,6 @@ gnutls_openpgp_crt_get_preferred_key_id (gnutls_openpgp_crt_t key,
  * This allows setting a preferred key id for the given certificate.
  * This key will be used by functions that involve key handling.
  *
- * If the provided @keyid is %NULL then the master key is
- * set as preferred.
- *
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned,
  *   otherwise a negative error code is returned.
  **/
@@ -1605,20 +1566,6 @@ gnutls_openpgp_crt_set_preferred_key_id (gnutls_openpgp_crt_t key,
       gnutls_assert ();
       return GNUTLS_E_INVALID_REQUEST;
     }
-    
-  if (keyid == NULL) /* set the master as preferred */
-    {
-      uint8_t tmp[GNUTLS_OPENPGP_KEYID_SIZE];
-      
-      ret = gnutls_openpgp_crt_get_key_id (key, tmp);
-      if (ret < 0)
-        return gnutls_assert_val(ret);
-        
-      key->preferred_set = 1;
-      memcpy (key->preferred_keyid, tmp, GNUTLS_OPENPGP_KEYID_SIZE);
-
-      return 0;
-    }
 
   /* check if the id is valid */
   ret = gnutls_openpgp_crt_get_subkey_idx (key, keyid);
@@ -1639,10 +1586,14 @@ gnutls_openpgp_crt_set_preferred_key_id (gnutls_openpgp_crt_t key,
  * gnutls_openpgp_crt_get_auth_subkey:
  * @crt: the structure that contains the OpenPGP public key.
  * @keyid: the struct to save the keyid.
- * @flag: Non-zero indicates that a valid subkey is always returned.
+ * @flag: Non (0) indicates that a valid subkey is always returned.
  *
  * Returns the 64-bit keyID of the first valid OpenPGP subkey marked
- * for authentication.  If flag is non-zero and no authentication
+ * for authentication.  If flag is non (0) and no authentication
+ * subkey exists, then a valid subkey will be returned even if it is
+ * not marked for authentication.
+ * Returns the 64-bit keyID of the first valid OpenPGP subkey marked
+ * for authentication.  If flag is non (0) and no authentication
  * subkey exists, then a valid subkey will be returned even if it is
  * not marked for authentication.
  *
