@@ -30,7 +30,7 @@
 #include <gnutls_errors.h>
 #include <gnutls_datum.h>
 
-
+#include <pin.h>
 #include <pkcs11_int.h>
 #include <p11-kit/p11-kit.h>
 #include <p11-kit/pin.h>
@@ -39,11 +39,12 @@
 
 /* XXX: try to eliminate this */
 #define MAX_CERT_SIZE 8*1024
-#define MAX_SLOTS 48
 
 struct gnutls_pkcs11_provider_s
 {
   struct ck_function_list *module;
+  unsigned long nslots;
+  ck_slot_id_t *slots;
   struct ck_info info;
   unsigned int initialized;
 };
@@ -73,11 +74,8 @@ static struct gnutls_pkcs11_provider_s providers[MAX_PROVIDERS];
 static unsigned int active_providers = 0;
 static unsigned int initialized_registered = 0;
 
-static gnutls_pkcs11_pin_callback_t pin_func;
-static void *pin_data;
-
-gnutls_pkcs11_token_callback_t token_func;
-void *token_data;
+gnutls_pkcs11_token_callback_t _gnutls_token_func;
+void *_gnutls_token_data;
 
 int
 pkcs11_rv_to_err (ck_rv_t rv)
@@ -166,19 +164,14 @@ pkcs11_rv_to_err (ck_rv_t rv)
     }
 }
 
-
-static int scan_slots(struct gnutls_pkcs11_provider_s * p, ck_slot_id_t *slots,
-	 unsigned long *nslots)
+/* Fake scan */
+void
+pkcs11_rescan_slots (void)
 {
-ck_rv_t rv;
+  unsigned long slots;
 
-  rv = pkcs11_get_slot_list(p->module, 1, slots, nslots);
-  if (rv != CKR_OK)
-    {
-      gnutls_assert ();
-      return pkcs11_rv_to_err(rv);
-    }
-  return 0;
+  pkcs11_get_slot_list (providers[active_providers - 1].module, 0,
+                          NULL, &slots);
 }
 
 static int
@@ -209,11 +202,44 @@ pkcs11_add_module (const char *name, struct ck_function_list *module)
   active_providers++;
   providers[active_providers - 1].module = module;
 
+  /* cache the number of slots in this module */
+  if (pkcs11_get_slot_list
+      (providers[active_providers - 1].module, 0, NULL,
+       &providers[active_providers - 1].nslots) != CKR_OK)
+    {
+      gnutls_assert ();
+      goto fail;
+    }
+
+  providers[active_providers - 1].slots =
+    gnutls_malloc (sizeof (*providers[active_providers - 1].slots) *
+                   providers[active_providers - 1].nslots);
+  if (providers[active_providers - 1].slots == NULL)
+    {
+      gnutls_assert ();
+      goto fail;
+    }
+
+  if (pkcs11_get_slot_list
+      (providers[active_providers - 1].module, 0,
+       providers[active_providers - 1].slots,
+       &providers[active_providers - 1].nslots) != CKR_OK)
+    {
+      gnutls_assert ();
+      gnutls_free (providers[active_providers - 1].slots);
+      goto fail;
+    }
+
   memcpy (&providers[active_providers - 1].info, &info, sizeof(info));
 
-  _gnutls_debug_log ("p11: loaded provider '%s'\n", name);
+  _gnutls_debug_log ("p11: loaded provider '%s' with %d slots\n",
+                     name, (int) providers[active_providers - 1].nslots);
 
   return 0;
+
+fail:
+  active_providers--;
+  return GNUTLS_E_PKCS11_LOAD_ERROR;
 }
 
 
@@ -237,10 +263,12 @@ gnutls_pkcs11_add_provider (const char *name, const char *params)
   struct ck_function_list *module;
   int ret;
 
+  active_providers++;
   if (p11_kit_load_initialize_module (name, &module) != CKR_OK)
     {
       gnutls_assert ();
       _gnutls_debug_log ("p11: Cannot load provider %s\n", name);
+      active_providers--;
       return GNUTLS_E_PKCS11_LOAD_ERROR;
     }
 
@@ -472,7 +500,7 @@ initialize_automatic_p11_kit (void)
       gnutls_assert ();
       _gnutls_debug_log ("Cannot initialize registered module: %s\n",
                          p11_kit_strerror (rv));
-      return pkcs11_rv_to_err(rv);
+      return GNUTLS_E_INTERNAL_ERROR;
     }
 
   initialized_registered = 1;
@@ -558,18 +586,15 @@ gnutls_pkcs11_init (unsigned int flags, const char *deprecated_config_file)
  **/
 int gnutls_pkcs11_reinit (void)
 {
-  unsigned i;
   int rv;
 
-  for (i = 0; i < active_providers; i++)
+  rv = p11_kit_initialize_registered ();
+  if (rv != CKR_OK)
     {
-      if (providers[i].module != NULL) 
-        {
-          rv = p11_kit_initialize_module(providers[i].module);
-          if (rv != CKR_OK)
-            _gnutls_debug_log ("Cannot initialize registered module '%s': %s\n", 
-              providers[i].info.library_description, p11_kit_strerror (rv));
-        }
+      gnutls_assert ();
+      _gnutls_debug_log ("Cannot initialize registered module: %s\n",
+                         p11_kit_strerror (rv));
+      return GNUTLS_E_INTERNAL_ERROR;
     }
 
   return 0;
@@ -602,29 +627,13 @@ gnutls_pkcs11_deinit (void)
         p11_kit_finalize_module (providers[i].module);
     }
   active_providers = 0;
-
+  
   if (initialized_registered != 0)
     p11_kit_finalize_registered ();
   initialized_registered = 0;
-}
-
-/**
- * gnutls_pkcs11_set_pin_function:
- * @fn: The PIN callback, a gnutls_pkcs11_pin_callback_t() function.
- * @userdata: data to be supplied to callback
- *
- * This function will set a callback function to be used when a PIN is
- * required for PKCS 11 operations.  See
- * gnutls_pkcs11_pin_callback_t() on how the callback should behave.
- *
- * Since: 2.12.0
- **/
-void
-gnutls_pkcs11_set_pin_function (gnutls_pkcs11_pin_callback_t fn,
-                                void *userdata)
-{
-  pin_func = fn;
-  pin_data = userdata;
+  
+  gnutls_pkcs11_set_pin_function (NULL, NULL);
+  gnutls_pkcs11_set_token_function (NULL, NULL);
 }
 
 /**
@@ -641,8 +650,8 @@ void
 gnutls_pkcs11_set_token_function (gnutls_pkcs11_token_callback_t fn,
                                   void *userdata)
 {
-  token_func = fn;
-  token_data = userdata;
+  _gnutls_token_func = fn;
+  _gnutls_token_data = userdata;
 }
 
 int
@@ -741,6 +750,26 @@ gnutls_pkcs11_obj_init (gnutls_pkcs11_obj_t * obj)
   return 0;
 }
 
+/**
+ * gnutls_pkcs11_obj_set_pin_function:
+ * @obj: The object structure
+ * @fn: the callback
+ * @userdata: data associated with the callback
+ *
+ * This function will set a callback function to be used when
+ * required to access the object. This function overrides the global
+ * set using gnutls_pkcs11_set_pin_function().
+ *
+ * Since: 3.1.0
+ *
+ **/
+void gnutls_pkcs11_obj_set_pin_function (gnutls_pkcs11_obj_t obj,
+                                         gnutls_pin_callback_t fn, void *userdata)
+{
+  obj->pin.cb = fn;
+  obj->pin.data = userdata;
+}
+
 /**
  * gnutls_pkcs11_obj_deinit:
  * @obj: The structure to be initialized
@@ -803,21 +832,19 @@ gnutls_pkcs11_obj_export (gnutls_pkcs11_obj_t obj,
 }
 
 int
-pkcs11_find_object (struct ck_function_list ** _module,
-                    ck_session_handle_t * _pks,
+pkcs11_find_object (struct pkcs11_session_info* sinfo,
+                    struct pin_info_st * pin_info,
                     ck_object_handle_t * _obj,
                     struct p11_kit_uri *info, unsigned int flags)
 {
   int ret;
-  struct ck_function_list *module;
-  ck_session_handle_t pks;
   ck_object_handle_t obj;
   struct ck_attribute *attrs;
   unsigned long attr_count;
   unsigned long count;
   ck_rv_t rv;
 
-  ret = pkcs11_open_session (&module, &pks, info, flags & SESSION_LOGIN);
+  ret = pkcs11_open_session (sinfo, pin_info, info, flags & SESSION_LOGIN);
   if (ret < 0)
     {
       gnutls_assert ();
@@ -825,7 +852,7 @@ pkcs11_find_object (struct ck_function_list ** _module,
     }
 
   attrs = p11_kit_uri_get_attributes (info, &attr_count);
-  rv = pkcs11_find_objects_init (module, pks, attrs, attr_count);
+  rv = pkcs11_find_objects_init (sinfo->module, sinfo->pks, attrs, attr_count);
   if (rv != CKR_OK)
     {
       gnutls_assert ();
@@ -834,19 +861,17 @@ pkcs11_find_object (struct ck_function_list ** _module,
       goto fail;
     }
 
-  if (pkcs11_find_objects (module, pks, &obj, 1, &count) == CKR_OK && count == 1)
+  if (pkcs11_find_objects (sinfo->module, sinfo->pks, &obj, 1, &count) == CKR_OK && count == 1)
     {
       *_obj = obj;
-      *_pks = pks;
-      *_module = module;
-      pkcs11_find_objects_final (module, pks);
+      pkcs11_find_objects_final (sinfo);
       return 0;
     }
 
   ret = GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE;
-  pkcs11_find_objects_final (module, pks);
+  pkcs11_find_objects_final (sinfo);
 fail:
-  pkcs11_close_session (module, pks);
+  pkcs11_close_session (sinfo);
 
   return ret;
 }
@@ -856,34 +881,24 @@ pkcs11_find_slot (struct ck_function_list ** module, ck_slot_id_t * slot,
                   struct p11_kit_uri *info, struct token_info *_tinfo)
 {
   unsigned int x, z;
-  int ret;
-  unsigned long nslots;
-  ck_slot_id_t slots[MAX_SLOTS];
 
   for (x = 0; x < active_providers; x++)
     {
-      nslots = sizeof(slots)/sizeof(slots[0]);
-      ret = scan_slots(&providers[x], slots, &nslots);
-      if (ret < 0)
-        {
-          gnutls_assert();
-          continue;
-        }
-
-      for (z = 0; z < nslots; z++)
+      for (z = 0; z < providers[x].nslots; z++)
         {
           struct token_info tinfo;
 
           if (pkcs11_get_token_info
-              (providers[x].module, slots[z], &tinfo.tinfo) != CKR_OK)
+              (providers[x].module, providers[x].slots[z],
+               &tinfo.tinfo) != CKR_OK)
             {
               continue;
             }
-          tinfo.sid = slots[z];
+          tinfo.sid = providers[x].slots[z];
           tinfo.prov = &providers[x];
 
           if (pkcs11_get_slot_info
-              (providers[x].module, slots[z],
+              (providers[x].module, providers[x].slots[z],
                &tinfo.sinfo) != CKR_OK)
             {
               continue;
@@ -897,7 +912,7 @@ pkcs11_find_slot (struct ck_function_list ** module, ck_slot_id_t * slot,
 
           /* ok found */
           *module = providers[x].module;
-          *slot = slots[z];
+          *slot = providers[x].slots[z];
 
           if (_tinfo != NULL)
             memcpy (_tinfo, &tinfo, sizeof (tinfo));
@@ -911,8 +926,10 @@ pkcs11_find_slot (struct ck_function_list ** module, ck_slot_id_t * slot,
 }
 
 int
-pkcs11_open_session (struct ck_function_list ** _module, ck_session_handle_t * _pks,
-                     struct p11_kit_uri *info, unsigned int flags)
+pkcs11_open_session (struct pkcs11_session_info *sinfo, 
+                     struct pin_info_st *pin_info,
+                     struct p11_kit_uri *info, 
+                     unsigned int flags)
 {
   ck_rv_t rv;
   int ret;
@@ -938,66 +955,64 @@ pkcs11_open_session (struct ck_function_list ** _module, ck_session_handle_t * _
       return pkcs11_rv_to_err (rv);
     }
 
+  /* ok found */
+  sinfo->pks = pks;
+  sinfo->module = module;
+  sinfo->init = 1;
+  memcpy(&sinfo->tinfo, &tinfo.tinfo, sizeof(sinfo->tinfo));
+
   if (flags & SESSION_LOGIN)
     {
-      ret = pkcs11_login (module, pks, &tinfo, info, (flags & SESSION_SO) ? 1 : 0);
+      ret = pkcs11_login (sinfo, pin_info, &tinfo, info, (flags & SESSION_SO) ? 1 : 0);
       if (ret < 0)
         {
           gnutls_assert ();
-          pkcs11_close_session (module, pks);
+          pkcs11_close_session (sinfo);
           return ret;
         }
     }
 
-  /* ok found */
-  *_pks = pks;
-  *_module = module;
   return 0;
 }
 
 
 int
 _pkcs11_traverse_tokens (find_func_t find_func, void *input,
-                         struct p11_kit_uri *info, unsigned int flags)
+                         struct p11_kit_uri *info, 
+                         struct pin_info_st *pin_info,
+                         unsigned int flags)
 {
   ck_rv_t rv;
   unsigned int found = 0, x, z;
   int ret;
   ck_session_handle_t pks = 0;
+  struct pkcs11_session_info sinfo;
   struct ck_function_list *module = NULL;
-  unsigned long nslots;
-  ck_slot_id_t slots[MAX_SLOTS];
 
   for (x = 0; x < active_providers; x++)
     {
-      nslots = sizeof(slots)/sizeof(slots[0]);
-      ret = scan_slots(&providers[x], slots, &nslots);
-      if (ret < 0)
-        {
-          gnutls_assert();
-          continue;
-        }
-
       module = providers[x].module;
-      for (z = 0; z < nslots; z++)
+      for (z = 0; z < providers[x].nslots; z++)
         {
           struct token_info tinfo;
 
-          if (pkcs11_get_token_info (module, slots[z],
+          ret = GNUTLS_E_PKCS11_ERROR;
+
+          if (pkcs11_get_token_info (module, providers[x].slots[z],
                &tinfo.tinfo) != CKR_OK)
             {
               continue;
             }
-          tinfo.sid = slots[z];
+          tinfo.sid = providers[x].slots[z];
           tinfo.prov = &providers[x];
 
-          if (pkcs11_get_slot_info (module, slots[z],
+          if (pkcs11_get_slot_info (module, providers[x].slots[z],
                &tinfo.sinfo) != CKR_OK)
             {
               continue;
             }
 
-          rv = (module)->C_OpenSession (slots[z],
+          rv = (module)->C_OpenSession (providers[x].slots[z],
                                         ((flags & SESSION_WRITE)
                                           ? CKF_RW_SESSION : 0) |
                                         CKF_SERIAL_SESSION, NULL, NULL, &pks);
@@ -1005,10 +1020,13 @@ _pkcs11_traverse_tokens (find_func_t find_func, void *input,
             {
               continue;
             }
+          
+          sinfo.module = module;
+          sinfo.pks = pks;
 
           if (flags & SESSION_LOGIN)
             {
-              ret = pkcs11_login (module, pks, &tinfo, info, (flags & SESSION_SO) ? 1 : 0);
+              ret = pkcs11_login (&sinfo, pin_info, &tinfo, info, (flags & SESSION_SO) ? 1 : 0);
               if (ret < 0)
                 {
                   gnutls_assert ();
@@ -1016,7 +1034,7 @@ _pkcs11_traverse_tokens (find_func_t find_func, void *input,
                 }
             }
 
-          ret = find_func (module, pks, &tinfo, &providers[x].info, input);
+          ret = find_func (&sinfo, &tinfo, &providers[x].info, input);
 
           if (ret == 0)
             {
@@ -1025,7 +1043,7 @@ _pkcs11_traverse_tokens (find_func_t find_func, void *input,
             }
           else
             {
-              pkcs11_close_session (module, pks);
+              pkcs11_close_session (&sinfo);
               pks = 0;
             }
         }
@@ -1037,7 +1055,11 @@ finish:
   if (found == 0)
     {
       if (module)
-        ret = find_func (module, pks, NULL, NULL, input);
+        {
+          sinfo.module = module;
+          sinfo.pks = pks;
+          ret = find_func (&sinfo, NULL, NULL, input);
+        }
       else
         ret = gnutls_assert_val(GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE);
     }
@@ -1048,7 +1070,7 @@ finish:
 
   if (pks != 0 && module != NULL)
     {
-      pkcs11_close_session (module, pks);
+      pkcs11_close_session (&sinfo);
     }
 
   return ret;
@@ -1391,7 +1413,7 @@ pkcs11_obj_import_pubkey (struct ck_function_list *module,
 }
 
 static int
-find_obj_url (struct ck_function_list *module, ck_session_handle_t pks,
+find_obj_url (struct pkcs11_session_info *sinfo,
               struct token_info *info, struct ck_info *lib_info, void *input)
 {
   struct url_find_data_st *find_data = input;
@@ -1472,7 +1494,7 @@ find_obj_url (struct ck_function_list *module, ck_session_handle_t pks,
       a_vals++;
     }
 
-  rv = pkcs11_find_objects_init (module, pks, a, a_vals);
+  rv = pkcs11_find_objects_init (sinfo->module, sinfo->pks, a, a_vals);
   if (rv != CKR_OK)
     {
       gnutls_assert ();
@@ -1481,7 +1503,7 @@ find_obj_url (struct ck_function_list *module, ck_session_handle_t pks,
       goto cleanup;
     }
 
-  while (pkcs11_find_objects (module, pks, &obj, 1, &count) == CKR_OK && count == 1)
+  while (pkcs11_find_objects (sinfo->module, sinfo->pks, &obj, 1, &count) == CKR_OK && count == 1)
     {
 
       a[0].type = CKA_VALUE;
@@ -1494,7 +1516,7 @@ find_obj_url (struct ck_function_list *module, ck_session_handle_t pks,
       a[2].value = id_tmp;
       a[2].value_len = sizeof(id_tmp);
 
-      if (pkcs11_get_attribute_value (module, pks, obj, a, 3) == CKR_OK)
+      if (pkcs11_get_attribute_value (sinfo->module, sinfo->pks, obj, a, 3) == CKR_OK)
         {
           gnutls_datum_t id = { a[2].value, a[2].value_len };
           gnutls_datum_t data = { a[0].value, a[0].value_len };
@@ -1503,7 +1525,7 @@ find_obj_url (struct ck_function_list *module, ck_session_handle_t pks,
           if (class == CKO_PUBLIC_KEY)
             {
               ret =
-                pkcs11_obj_import_pubkey (module, pks, obj,
+                pkcs11_obj_import_pubkey (sinfo->module, sinfo->pks, obj,
                                           find_data->crt,
                                           &id, &label,
                                           &info->tinfo, lib_info);
@@ -1543,7 +1565,7 @@ find_obj_url (struct ck_function_list *module, ck_session_handle_t pks,
 
 cleanup:
   gnutls_free (cert_data);
-  pkcs11_find_objects_final (module, pks);
+  pkcs11_find_objects_final (sinfo);
 
   return ret;
 }
@@ -1563,12 +1585,12 @@ pkcs11_obj_flags_to_int (unsigned int flags)
 
 /**
  * gnutls_pkcs11_obj_import_url:
- * @cert: The structure to store the parsed certificate
+ * @obj: The structure to store the object
  * @url: a PKCS 11 url identifying the key
  * @flags: One of GNUTLS_PKCS11_OBJ_* flags
  *
- * This function will "import" a PKCS 11 URL identifying a certificate
- * key to the #gnutls_pkcs11_obj_t structure. This does not involve any
+ * This function will "import" a PKCS 11 URL identifying an object (e.g. certificate)
+ * to the #gnutls_pkcs11_obj_t structure. This does not involve any
  * parsing (such as X.509 or OpenPGP) since the #gnutls_pkcs11_obj_t is
  * format agnostic. Only data are transferred.
  *
@@ -1578,16 +1600,16 @@ pkcs11_obj_flags_to_int (unsigned int flags)
  * Since: 2.12.0
  **/
 int
-gnutls_pkcs11_obj_import_url (gnutls_pkcs11_obj_t cert, const char *url,
+gnutls_pkcs11_obj_import_url (gnutls_pkcs11_obj_t obj, const char *url,
                               unsigned int flags)
 {
   int ret;
   struct url_find_data_st find_data;
 
   /* fill in the find data structure */
-  find_data.crt = cert;
+  find_data.crt = obj;
 
-  ret = pkcs11_url_to_info (url, &cert->info);
+  ret = pkcs11_url_to_info (url, &obj->info);
   if (ret < 0)
     {
       gnutls_assert ();
@@ -1595,8 +1617,8 @@ gnutls_pkcs11_obj_import_url (gnutls_pkcs11_obj_t cert, const char *url,
     }
 
   ret =
-    _pkcs11_traverse_tokens (find_obj_url, &find_data, cert->info,
-                             pkcs11_obj_flags_to_int (flags));
+    _pkcs11_traverse_tokens (find_obj_url, &find_data, obj->info,
+                             &obj->pin, pkcs11_obj_flags_to_int (flags));
 
   if (ret < 0)
     {
@@ -1615,8 +1637,7 @@ struct token_num
 };
 
 static int
-find_token_num (struct ck_function_list *module,
-                ck_session_handle_t pks,
+find_token_num (struct pkcs11_session_info* sinfo,
                 struct token_info *tinfo,
                 struct ck_info *lib_info, void *input)
 {
@@ -1668,7 +1689,7 @@ gnutls_pkcs11_token_get_url (unsigned int seq,
   tn.seq = seq;
   tn.info = p11_kit_uri_new ();
 
-  ret = _pkcs11_traverse_tokens (find_token_num, &tn, NULL, 0);
+  ret = _pkcs11_traverse_tokens (find_token_num, &tn, NULL, NULL, 0);
   if (ret < 0)
     {
       p11_kit_uri_free (tn.info);
@@ -1881,8 +1902,10 @@ retrieve_pin_from_source (const char *pinfile, struct ck_token_info *token_info,
 }
 
 static int
-retrieve_pin_for_callback (struct ck_token_info *token_info, int attempts,
-                           ck_user_type_t user_type, struct p11_kit_pin **pin)
+retrieve_pin_from_callback (const struct pin_info_st *pin_info,
+                            struct ck_token_info *token_info, 
+                            int attempts, ck_user_type_t user_type, 
+                            struct p11_kit_pin **pin)
 {
   char pin_value[GNUTLS_PKCS11_MAX_PIN_LEN];
   unsigned int flags = 0;
@@ -1920,26 +1943,33 @@ retrieve_pin_for_callback (struct ck_token_info *token_info, int attempts,
 
   if (user_type == CKU_USER)
     {
-      flags |= GNUTLS_PKCS11_PIN_USER;
+      flags |= GNUTLS_PIN_USER;
       if (token_info->flags & CKF_USER_PIN_COUNT_LOW)
-        flags |= GNUTLS_PKCS11_PIN_COUNT_LOW;
+        flags |= GNUTLS_PIN_COUNT_LOW;
       if (token_info->flags & CKF_USER_PIN_FINAL_TRY)
-        flags |= GNUTLS_PKCS11_PIN_FINAL_TRY;
+        flags |= GNUTLS_PIN_FINAL_TRY;
     }
   else if (user_type == CKU_SO)
     {
-      flags |= GNUTLS_PKCS11_PIN_SO;
+      flags |= GNUTLS_PIN_SO;
       if (token_info->flags & CKF_SO_PIN_COUNT_LOW)
-        flags |= GNUTLS_PKCS11_PIN_COUNT_LOW;
+        flags |= GNUTLS_PIN_COUNT_LOW;
       if (token_info->flags & CKF_SO_PIN_FINAL_TRY)
-        flags |= GNUTLS_PKCS11_PIN_FINAL_TRY;
+        flags |= GNUTLS_PIN_FINAL_TRY;
     }
 
   if (attempts > 0)
-    flags |= GNUTLS_PKCS11_PIN_WRONG;
+    flags |= GNUTLS_PIN_WRONG;
+
+  if (pin_info && pin_info->cb)
+    ret = pin_info->cb (pin_info->data, attempts, (char*)token_str, label,
+                        flags, pin_value, GNUTLS_PKCS11_MAX_PIN_LEN);
+  else if (_gnutls_pin_func)
+    ret = _gnutls_pin_func (_gnutls_pin_data, attempts, (char*)token_str, label,
+                            flags, pin_value, GNUTLS_PKCS11_MAX_PIN_LEN);
+  else
+    ret = gnutls_assert_val(GNUTLS_E_PKCS11_PIN_ERROR);
 
-  ret = pin_func (pin_data, attempts, (char*)token_str, label,
-                  flags, pin_value, GNUTLS_PKCS11_MAX_PIN_LEN);
   free (token_str);
   free (label);
 
@@ -1955,8 +1985,9 @@ retrieve_pin_for_callback (struct ck_token_info *token_info, int attempts,
 }
 
 static int
-retrieve_pin (struct p11_kit_uri *info, struct ck_token_info *token_info,
-              int attempts, ck_user_type_t user_type, struct p11_kit_pin **pin)
+retrieve_pin (struct pin_info_st* pin_info, struct p11_kit_uri *info, 
+              struct ck_token_info *token_info, int attempts, 
+              ck_user_type_t user_type, struct p11_kit_pin **pin)
 {
   const char *pinfile;
   int ret = GNUTLS_E_PKCS11_PIN_ERROR;
@@ -1972,8 +2003,8 @@ retrieve_pin (struct p11_kit_uri *info, struct ck_token_info *token_info,
     }
 
   /* The global gnutls pin callback */
-  if (pin_func && ret < 0)
-    ret = retrieve_pin_for_callback (token_info, attempts, user_type, pin);
+  if (ret < 0)
+    ret = retrieve_pin_from_callback (pin_info, token_info, attempts, user_type, pin);
 
   /* Otherwise, PIN entry is necessary for login, so fail if there's
    * no callback. */
@@ -1988,7 +2019,7 @@ retrieve_pin (struct p11_kit_uri *info, struct ck_token_info *token_info,
 }
 
 int
-pkcs11_login (struct ck_function_list * module, ck_session_handle_t pks,
+pkcs11_login (struct pkcs11_session_info * sinfo, struct pin_info_st * pin_info,
               const struct token_info *tokinfo, struct p11_kit_uri *info, int so)
 {
   struct ck_session_info session_info;
@@ -2009,7 +2040,7 @@ pkcs11_login (struct ck_function_list * module, ck_session_handle_t pks,
    * required. */
   if (tokinfo->tinfo.flags & CKF_PROTECTED_AUTHENTICATION_PATH)
     {
-      rv = (module)->C_Login (pks, (so == 0) ? CKU_USER : CKU_SO, NULL, 0);
+      rv = (sinfo->module)->C_Login (sinfo->pks, (so == 0) ? CKU_USER : CKU_SO, NULL, 0);
       if (rv == CKR_OK || rv == CKR_USER_ALREADY_LOGGED_IN)
         {
           return 0;
@@ -2031,7 +2062,7 @@ pkcs11_login (struct ck_function_list * module, ck_session_handle_t pks,
       memcpy (&tinfo, &tokinfo->tinfo, sizeof(tinfo));
 
       /* Check whether the session is already logged in, and if so, just skip */
-      rv = (module)->C_GetSessionInfo (pks, &session_info);
+      rv = (sinfo->module)->C_GetSessionInfo (sinfo->pks, &session_info);
       if (rv == CKR_OK && (session_info.state == CKS_RO_USER_FUNCTIONS ||
                            session_info.state == CKS_RW_USER_FUNCTIONS))
         {
@@ -2053,14 +2084,14 @@ pkcs11_login (struct ck_function_list * module, ck_session_handle_t pks,
             }
         }
 
-      ret = retrieve_pin (info, &tinfo, attempt++, user_type, &pin);
+      ret = retrieve_pin (pin_info, info, &tinfo, attempt++, user_type, &pin);
       if (ret < 0)
         {
           gnutls_assert ();
           goto cleanup;
         }
 
-      rv = (module)->C_Login (pks, user_type,
+      rv = (sinfo->module)->C_Login (sinfo->pks, user_type,
                               (unsigned char *)p11_kit_pin_get_value (pin, NULL),
                               p11_kit_pin_get_length (pin));
 
@@ -2087,7 +2118,7 @@ pkcs11_call_token_func (struct p11_kit_uri *info, const unsigned retry)
 
   tinfo = p11_kit_uri_get_token_info (info);
   label = p11_kit_space_strdup (tinfo->label, sizeof (tinfo->label));
-  ret = (token_func) (token_data, label, retry);
+  ret = (_gnutls_token_func) (_gnutls_token_data, label, retry);
   free (label);
 
   return ret;
@@ -2095,7 +2126,7 @@ pkcs11_call_token_func (struct p11_kit_uri *info, const unsigned retry)
 
 
 static int
-find_privkeys (struct ck_function_list *module, ck_session_handle_t pks,
+find_privkeys (struct pkcs11_session_info* sinfo,
                struct token_info *info, struct pkey_list *list)
 {
   struct ck_attribute a[3];
@@ -2114,7 +2145,7 @@ find_privkeys (struct ck_function_list *module, ck_session_handle_t pks,
   a[0].value = &class;
   a[0].value_len = sizeof class;
 
-  rv = pkcs11_find_objects_init (module, pks, a, 1);
+  rv = pkcs11_find_objects_init (sinfo->module, sinfo->pks, a, 1);
   if (rv != CKR_OK)
     {
       gnutls_assert ();
@@ -2122,12 +2153,12 @@ find_privkeys (struct ck_function_list *module, ck_session_handle_t pks,
     }
 
   list->key_ids_size = 0;
-  while (pkcs11_find_objects (module, pks, &obj, 1, &count) == CKR_OK && count == 1)
+  while (pkcs11_find_objects (sinfo->module, sinfo->pks, &obj, 1, &count) == CKR_OK && count == 1)
     {
       list->key_ids_size++;
     }
 
-  pkcs11_find_objects_final (module, pks);
+  pkcs11_find_objects_final (sinfo);
 
   if (list->key_ids_size == 0)
     {
@@ -2148,7 +2179,7 @@ find_privkeys (struct ck_function_list *module, ck_session_handle_t pks,
   a[0].value = &class;
   a[0].value_len = sizeof class;
 
-  rv = pkcs11_find_objects_init (module, pks, a, 1);
+  rv = pkcs11_find_objects_init (sinfo->module, sinfo->pks, a, 1);
   if (rv != CKR_OK)
     {
       gnutls_assert ();
@@ -2156,7 +2187,7 @@ find_privkeys (struct ck_function_list *module, ck_session_handle_t pks,
     }
 
   current = 0;
-  while (pkcs11_find_objects (module, pks, &obj, 1, &count) == CKR_OK && count == 1)
+  while (pkcs11_find_objects (sinfo->module, sinfo->pks, &obj, 1, &count) == CKR_OK && count == 1)
     {
 
       a[0].type = CKA_ID;
@@ -2165,7 +2196,7 @@ find_privkeys (struct ck_function_list *module, ck_session_handle_t pks,
 
       _gnutls_buffer_init (&list->key_ids[current]);
 
-      if (pkcs11_get_attribute_value (module, pks, obj, a, 1) == CKR_OK)
+      if (pkcs11_get_attribute_value (sinfo->module, sinfo->pks, obj, a, 1) == CKR_OK)
         {
           _gnutls_buffer_append_data (&list->key_ids[current],
                                       a[0].value, a[0].value_len);
@@ -2176,7 +2207,7 @@ find_privkeys (struct ck_function_list *module, ck_session_handle_t pks,
         break;
     }
 
-  pkcs11_find_objects_final (module, pks);
+  pkcs11_find_objects_final (sinfo);
 
   list->key_ids_size = current - 1;
 
@@ -2187,16 +2218,15 @@ find_privkeys (struct ck_function_list *module, ck_session_handle_t pks,
 
 
 static int
-find_objs (struct ck_function_list * module, ck_session_handle_t pks,
+find_objs (struct pkcs11_session_info* sinfo,
            struct token_info *info, struct ck_info *lib_info, void *input)
 {
   struct crt_find_data_st *find_data = input;
-  struct ck_attribute a[6];
+  struct ck_attribute a[4];
   struct ck_attribute *attr;
   ck_object_class_t class = (ck_object_class_t)-1;
   ck_certificate_type_t type = (ck_certificate_type_t)-1;
-  unsigned char trusted;
-  unsigned long category;
+  unsigned int trusted;
   ck_rv_t rv;
   ck_object_handle_t obj;
   unsigned long count;
@@ -2232,7 +2262,7 @@ find_objs (struct ck_function_list * module, ck_session_handle_t pks,
 
   if (find_data->flags == GNUTLS_PKCS11_OBJ_ATTR_CRT_WITH_PRIVKEY)
     {
-      ret = find_privkeys (module, pks, info, &plist);
+      ret = find_privkeys (sinfo, info, &plist);
       if (ret < 0)
         {
           gnutls_assert ();
@@ -2292,28 +2322,6 @@ find_objs (struct ck_function_list * module, ck_session_handle_t pks,
       tot_values++;
 
     }
-  else if (find_data->flags == GNUTLS_PKCS11_OBJ_ATTR_CRT_TRUSTED_CA)
-    {
-      class = CKO_CERTIFICATE;
-      type = CKC_X_509;
-      trusted = 1;
-
-      a[tot_values].type = CKA_CLASS;
-      a[tot_values].value = &class;
-      a[tot_values].value_len = sizeof class;
-      tot_values++;
-
-      a[tot_values].type = CKA_TRUSTED;
-      a[tot_values].value = &trusted;
-      a[tot_values].value_len = sizeof trusted;
-      tot_values++;
-
-      category = 2;
-      a[tot_values].type = CKA_CERTIFICATE_CATEGORY;
-      a[tot_values].value = &category;
-      a[tot_values].value_len = sizeof category;
-      tot_values++;
-    }
   else if (find_data->flags == GNUTLS_PKCS11_OBJ_ATTR_PUBKEY)
     {
       class = CKO_PUBLIC_KEY;
@@ -2363,7 +2371,7 @@ find_objs (struct ck_function_list * module, ck_session_handle_t pks,
       tot_values++;
     }
 
-  rv = pkcs11_find_objects_init (module, pks, a, tot_values);
+  rv = pkcs11_find_objects_init (sinfo->module, sinfo->pks, a, tot_values);
   if (rv != CKR_OK)
     {
       gnutls_assert ();
@@ -2371,7 +2379,7 @@ find_objs (struct ck_function_list * module, ck_session_handle_t pks,
       return pkcs11_rv_to_err (rv);
     }
 
-  while (pkcs11_find_objects (module, pks, &obj, 1, &count) == CKR_OK && count == 1)
+  while (pkcs11_find_objects (sinfo->module, sinfo->pks, &obj, 1, &count) == CKR_OK && count == 1)
     {
       gnutls_datum_t label, id, value;
 
@@ -2379,7 +2387,7 @@ find_objs (struct ck_function_list * module, ck_session_handle_t pks,
       a[0].value = label_tmp;
       a[0].value_len = sizeof label_tmp;
 
-      if (pkcs11_get_attribute_value (module, pks, obj, a, 1) == CKR_OK)
+      if (pkcs11_get_attribute_value (sinfo->module, sinfo->pks, obj, a, 1) == CKR_OK)
         {
           label.data = a[0].value;
           label.size = a[0].value_len;
@@ -2394,7 +2402,7 @@ find_objs (struct ck_function_list * module, ck_session_handle_t pks,
       a[0].value = certid_tmp;
       a[0].value_len = sizeof certid_tmp;
 
-      if (pkcs11_get_attribute_value (module, pks, obj, a, 1) == CKR_OK)
+      if (pkcs11_get_attribute_value (sinfo->module, sinfo->pks, obj, a, 1) == CKR_OK)
         {
           id.data = a[0].value;
           id.size = a[0].value_len;
@@ -2408,7 +2416,7 @@ find_objs (struct ck_function_list * module, ck_session_handle_t pks,
       a[0].type = CKA_VALUE;
       a[0].value = cert_data;
       a[0].value_len = MAX_CERT_SIZE;
-      if (pkcs11_get_attribute_value (module, pks, obj, a, 1) == CKR_OK)
+      if (pkcs11_get_attribute_value (sinfo->module, sinfo->pks, obj, a, 1) == CKR_OK)
         {
           value.data = a[0].value;
           value.size = a[0].value_len;
@@ -2425,7 +2433,7 @@ find_objs (struct ck_function_list * module, ck_session_handle_t pks,
           a[0].value = &class;
           a[0].value_len = sizeof class;
 
-          pkcs11_get_attribute_value (module, pks, obj, a, 1);
+          pkcs11_get_attribute_value (sinfo->module, sinfo->pks, obj, a, 1);
         }
 
       if (find_data->flags == GNUTLS_PKCS11_OBJ_ATTR_CRT_WITH_PRIVKEY)
@@ -2456,7 +2464,7 @@ find_objs (struct ck_function_list * module, ck_session_handle_t pks,
           if (class == CKO_PUBLIC_KEY)
             {
               ret =
-                pkcs11_obj_import_pubkey (module, pks, obj,
+                pkcs11_obj_import_pubkey (sinfo->module, sinfo->pks, obj,
                                           find_data->p_list
                                           [find_data->current],
                                           &id, &label,
@@ -2483,13 +2491,13 @@ find_objs (struct ck_function_list * module, ck_session_handle_t pks,
     }
 
   gnutls_free (cert_data);
-  pkcs11_find_objects_final (module, pks);
+  pkcs11_find_objects_final (sinfo);
 
   return GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE; /* continue until all tokens have been checked */
 
 fail:
   gnutls_free (cert_data);
-  pkcs11_find_objects_final (module, pks);
+  pkcs11_find_objects_final (sinfo);
   if (plist.key_ids != NULL)
     {
       for (i = 0; i < plist.key_ids_size; i++)
@@ -2555,23 +2563,76 @@ gnutls_pkcs11_obj_list_import_url (gnutls_pkcs11_obj_t * p_list,
 
   ret =
     _pkcs11_traverse_tokens (find_objs, &find_data, find_data.info,
+                             NULL,
                              pkcs11_obj_flags_to_int (flags));
   p11_kit_uri_free (find_data.info);
 
   if (ret < 0)
     {
       gnutls_assert ();
-      if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
-        {
-          *n_list = 0;
-          ret = 0;
-        }
       return ret;
     }
 
   return 0;
 }
 
+/**
+ * gnutls_pkcs11_obj_list_import_url2:
+ * @p_list: An uninitialized object list (may be NULL)
+ * @n_list: It will contain the size of the list.
+ * @url: A PKCS 11 url identifying a set of objects
+ * @attrs: Attributes of type #gnutls_pkcs11_obj_attr_t that can be used to limit output
+ * @flags: One of GNUTLS_PKCS11_OBJ_* flags
+ *
+ * This function will initialize and set values to an object list
+ * by using all objects identified by the PKCS 11 URL. The output
+ * is stored in @p_list, which will be initialized.
+ *
+ * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
+ *   negative error value.
+ *
+ * Since: 3.1.0
+ **/
+int
+gnutls_pkcs11_obj_list_import_url2 (gnutls_pkcs11_obj_t ** p_list,
+                                   unsigned int *n_list,
+                                   const char *url,
+                                   gnutls_pkcs11_obj_attr_t attrs,
+                                   unsigned int flags)
+{
+unsigned int init = 1024;
+int ret;
+
+  *p_list = gnutls_malloc(sizeof(gnutls_pkcs11_obj_t)*init);
+  if (*p_list == NULL)
+    {
+      gnutls_assert();
+      return GNUTLS_E_MEMORY_ERROR;
+    }
+
+  ret = gnutls_pkcs11_obj_list_import_url( *p_list, &init, url, attrs, flags);
+  if (ret == GNUTLS_E_SHORT_MEMORY_BUFFER)
+    {
+      *p_list = gnutls_realloc_fast(*p_list, sizeof(gnutls_pkcs11_obj_t)*init);
+      if (*p_list == NULL)
+        return gnutls_assert_val(GNUTLS_E_MEMORY_ERROR);
+      
+      ret = gnutls_pkcs11_obj_list_import_url( *p_list, &init, url, attrs, flags);
+    }
+
+  if (ret < 0)
+    {
+      gnutls_assert();
+      gnutls_free(*p_list);
+      *p_list = NULL;
+      return ret;
+    }
+
+  *n_list = init;
+  return 0;
+
+}
+
 /**
  * gnutls_x509_crt_import_pkcs11_url:
  * @crt: A certificate of type #gnutls_x509_crt_t
@@ -2600,6 +2661,9 @@ gnutls_x509_crt_import_pkcs11_url (gnutls_x509_crt_t crt,
       gnutls_assert ();
       return ret;
     }
+  
+  if (crt->pin.cb)
+    gnutls_pkcs11_obj_set_pin_function (pcrt, crt->pin.cb, crt->pin.data);
 
   ret = gnutls_pkcs11_obj_import_url (pcrt, url, flags);
   if (ret < 0)
@@ -2623,7 +2687,6 @@ cleanup:
   return ret;
 }
 
-
 /**
  * gnutls_x509_crt_import_pkcs11:
  * @crt: A certificate of type #gnutls_x509_crt_t
@@ -2697,7 +2760,7 @@ cleanup:
 }
 
 static int
-find_flags (struct ck_function_list * module, ck_session_handle_t pks,
+find_flags (struct pkcs11_session_info* sinfo,
             struct token_info *info, struct ck_info *lib_info, void *input)
 {
   struct flags_find_data_st *find_data = input;
@@ -2750,7 +2813,7 @@ gnutls_pkcs11_token_get_flags (const char *url, unsigned int *flags)
       return ret;
     }
 
-  ret = _pkcs11_traverse_tokens (find_flags, &find_data, find_data.info, 0);
+  ret = _pkcs11_traverse_tokens (find_flags, &find_data, find_data.info, NULL, 0);
   p11_kit_uri_free (find_data.info);
 
   if (ret < 0)
@@ -2915,17 +2978,16 @@ pkcs11_find_objects (struct ck_function_list *module,
 }
 
 ck_rv_t
-pkcs11_find_objects_final (struct ck_function_list *module,
-                           ck_session_handle_t sess)
+pkcs11_find_objects_final (struct pkcs11_session_info* sinfo)
 {
-	return (module)->C_FindObjectsFinal (sess);
+	return (sinfo->module)->C_FindObjectsFinal (sinfo->pks);
 }
 
 ck_rv_t
-pkcs11_close_session (struct ck_function_list *module,
-                      ck_session_handle_t sess)
+pkcs11_close_session (struct pkcs11_session_info * sinfo)
 {
-	return (module)->C_CloseSession (sess);
+        sinfo->init = 0;
+	return (sinfo->module)->C_CloseSession (sinfo->pks);
 }
 
 ck_rv_t
